from collections.abc import Callable

import venusian
from xcomponent import Catalog
from xcomponent.service.catalog import Component, Function

VENUSIAN_CATEGORY = "fastlife"  # copy/pasta from configurator


class XComponentRegistry:
    """
    Build the XComponent catalog using venusian.

    To avoid a global catalog, we generate the catalog by scanning using
    venusian.
    """

    components: dict[str, Component]

    functions: dict[str, Function]

    def __init__(self) -> None:
        self.components = {}
        self.functions = {}

    def register_xcomponent(self, name: str, component: Component) -> None:
        """Register a xcomponent component to the application template engine."""
        self.components[name] = component

    def register_xfunction(self, name: str, func: Function) -> None:
        """Register a xcomponent function to the application template engine."""
        self.functions[name] = func

    def build_catalog(self) -> Catalog:
        catalog = Catalog()
        for name, component in self.components.items():
            catalog.component(name)(component)

        for name, function in self.functions.items():
            catalog.function(name)(function)
        return catalog


def x_component(name: str | None = None) -> Callable[[Component], Component]:
    """
    Register a component to the XComponent catalog for the application.

    It differ from from @catalog.component decorator of the xcomponent library,
    in the way the catalog is created.
    The fastlife configurator will register the component to a catalog during
    the configuration.

    :param name: override the name of the component, default is the python function
        name.
    """
    component_name = name

    def decorator(wrapped: Component) -> Component:
        def callback(
            scanner: venusian.Scanner,
            name: str,
            ob: Component,
        ) -> None:
            if not hasattr(scanner, VENUSIAN_CATEGORY):
                return  # coverage: ignore
            scanner.fastlife.register_xcomponent(component_name or name, ob)  # type: ignore

        venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
        return wrapped

    return decorator


def x_function(name: str | None = None) -> Callable[[Function], Function]:
    """
    Register a component to the XComponent function for the application.

    It differ from from @catalog.function decorator of the xcomponent library,
    in the way the catalog is created.
    The fastlife configurator will register the function to a catalog during
    the configuration.

    :param name: override the name of the function, default is the python function name.
    """
    function_name = name

    def decorator(wrapped: Function) -> Function:
        def callback(
            scanner: venusian.Scanner,
            name: str,
            ob: Component,
        ) -> None:
            if not hasattr(scanner, VENUSIAN_CATEGORY):
                return  # coverage: ignore
            scanner.fastlife.register_xfunction(function_name or name, ob)  # type: ignore

        venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
        return wrapped

    return decorator



"""
Widget for field of type Union.
"""

from collections.abc import Sequence
from typing import TYPE_CHECKING, Union

from markupsafe import Markup
from pydantic import BaseModel, Field

if TYPE_CHECKING:
    from fastlife.service.templates import AbstractTemplateRenderer

from .base import TWidget, TypeWrapper, Widget


class UnionWidget(Widget[TWidget]):
    """
    Widget for union types.
    """

    template = """
    <Widget widget_id={id} removable={removable}>
      <div id={id}>
        <Details>
          <Summary id={id + '-union-summary'}>
            <H3 class={globals.H3_SUMMARY_CLASS}>{ globals.gettext(title) }</H3>
            <OptionalErrorText text={error} />
          </Summary>
          <div hx-sync="this" id={"id" + child}>
            {
              if child {
                child
              }
              else {
                for typ in types {
                  <Button type="button"
                    hx-target="closest div"
                    hx-get={typ.url}
                    hx-vals={typ.params}
                    id={typ.id}
                    onclick={"document.getElementById('" + id + "-remove-btn').hidden=false"}
                    class={globals.SECONDARY_BUTTON_CLASS}>{globals.gettext(typ.title)}</Button>
                }
              }
            }
          </div>
          <Button type="button" id={id + '-remove-btn'} hx-target={'#' + id}
            hx-vals={parent_type.params} hx-get={parent_type.url} hidden={not child}
            class={globals.SECONDARY_BUTTON_CLASS}>
            Remove
          </Button>
        </Details>
      </div>
    </Widget>
    """

    children_types: Sequence[type[BaseModel]]
    parent_type: TypeWrapper | None = Field(default=None)

    types: Sequence[TypeWrapper] | None = Field(default=None)
    child: str = Field(default="")

    def build_types(self, route_prefix: str) -> Sequence[TypeWrapper]:
        """Wrap types in the union in order to get the in their own widgets."""
        return [
            TypeWrapper(typ, route_prefix, self.name, self.token)
            for typ in self.children_types
        ]

    def to_html(self, renderer: "AbstractTemplateRenderer") -> Markup:
        """Return the html version."""
        self.child = Markup(self.value.to_html(renderer)) if self.value else ""
        self.types = self.build_types(renderer.route_prefix)
        self.parent_type = TypeWrapper(
            Union[tuple(self.children_types)],  # type: ignore # noqa: UP007
            renderer.route_prefix,
            self.name,
            self.token,
            title=self.title,
        )
        return Markup(renderer.render_template(self))



from typing import Any

from fastlife.adapters.xcomponent.registry import x_function


@x_function(name="len")
def length(sized: Any) -> int:
    return len(sized)


@x_function()
def is_bool(i: Any) -> bool:
    return isinstance(i, bool)


@x_function()
def is_str(value: Any) -> bool:
    return isinstance(value, str)
